/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (collect.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Mar 15 15:20:28 GMT 2000

  Updated:    Joerg Schulenburg 2004-2005

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: collect.C 

  PURPOSE: collect same configurations to reduce loops

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
// #include <math.h>
// #include <string.h>

#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"
// #include "../include/nr.h"
// #include "../include/nrutil.h"
#include "storage.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#ifndef DO_COLLECT
#define DO_COLLECT 1  /* set to 0 to switch off collect, default = 1 */
#endif

/* Joerg05:
 * who is calling  collect_term?
 *  62 times by obtain()+term2()..term62() in obtain.C 
 * parameters and its ranges?
 *  indices[0] - configuration (lattice animal), 1..NumberConfigurations
 *                                          new: 0 for GS Energy
 *  indices[1] - term 2..62, 999999=end (not used in solve!)
 *  indices[2..5] -  0,1..NumConfig ??  (0,0,0,0)=not_used(check it)
 * Outputfile fp1=*.input is used for solve()
 * Outputfile fp1=*_gs    is used for ...     (new Sep05) ???
 * maximum number of terms and clusters for sample models?
 *  MAX_NUMBER_TERMS_IN_CCM_EQNS to big? (*46Bytes)
 * ToDo: dynamic arrays
 */

#define MAX_out 1000

#if DO_COLLECT

// work: indices will be sorted (not used in the calling routine), Joerg Sep05
//       same or permutated indices[2..5] will be stored as sum only once
//       indices pointing to configurations 1..NumberConfigurations
//       permutated indices dont have the same probability! that means:
//       its difficult to simplify (no easy symmetry)
//       but it looks like that every term appears twice (ToDo: simplify!)
/* id: 0=gs_terms, 1=terms */
void collect_term(int indices[6], double dummy, int id) 
{
  int  k; long i, j; // , nio=7;
  static long n_out=0; /* count output terms, avoid GB of debugging output */
  static long n_in=0;  /* count input terms */
  static long na_in=0, na_out=0; /* summed over all bonds */
  static int  id_old=-1;         /* flag for new collection */
  static long counter=0;         /* buffer index */
  static long max_counter=0;     /* maximum buffer size */
  static int    descriptors[MAX_NUMBER_TERMS_IN_CCM_EQNS+1][6]; // 24 byte
  static double multipliers[MAX_NUMBER_TERMS_IN_CCM_EQNS+1];    //  8 byte
  // it is very slow to search a big table for an entry sequentialy
  // so we introduce a hash table of indizes to that entries
  //  and a table of indices to that table to have fast sequential access 
  //  to spare filled hash table (Joerg Sep05, xxzLC-14 5x faster)
  // number of hashed entries is in counter
  static long hash[MAX_NUMBER_TERMS_IN_CCM_EQNS+1]; // 8B hashed index to entries
  static int  hidx[MAX_NUMBER_TERMS_IN_CCM_EQNS+1]; // 4B index to hash entries
  static int hcol=0;  // counter for hash collisions, just as info
  static int hhit=0;  // counter for hash hits
  static int hnew=0;  // counter for new hash entries
  static int hval=-1;  // hash value
  static long numiterms[63], // count number of terms, only verbose output
              numoterms[63]; // ToDo: counting 1..4-fold-terms would be enough


  if (n_in==0 && id_old != id) {  // new file
    for (i=0;i<63;i++) numiterms[i]=numoterms[i]=0; // initialize counter
  }

  n_in++;
  if (hval<0) { hval=0; // used as flag, initialize hash once!
    for (i=0;i<MAX_NUMBER_TERMS_IN_CCM_EQNS+1;i++) { hidx[i]=hash[i]=0; }
  }

  if (counter){  // something stored into buffer
      /* save collected terms and reset buffer for next term */

    if (    descriptors[counter][0] < indices[0]  /* next configuration number? */
//  Joerg Sep05: indices[1]=term not used in solve, so we can collect it too
       || descriptors[counter][1] < indices[1]  /* next term? 2..62 */
       ||                       0 > indices[0]  /* flush buffer condition */
      /* Sept 05: old was 0>i[0] but collect_gse needs 0>=i[0] */
       ||    NumberConfigurations < indices[0]
       || counter >= MAX_NUMBER_TERMS_IN_CCM_EQNS) /* buffer full */
    {
        if (       MAX_NUMBER_TERMS_IN_CCM_EQNS == counter) {
           static long n_out2=0;
           if (n_out2<10) {
             printf("MAX_NUMBER_TERMS_IN_CCM_EQNS is to small, flush buffer\n");
             printf("Overflow happens for cfg= %d  term= %d (%d %d %d %d)\n",
               indices[0],indices[1],
               indices[2],indices[3],indices[4],indices[5]);
             fflush(stdout);
           }
           if (++n_out2==10)
            printf("excessive output is stopped after %d prints\n",MAX_out);
        }
	// first save information 
	for(j=1;j<=counter;j++){
	  
          term_store(id,descriptors[j],multipliers[j]);
	  
	  // printf out information
	  
	  if (Control.output > 1 && myid == 0 && n_out < MAX_out) {
	    printf("OUT: Config=%3d : Term = %2d : CCM = (%2d,%2d,%2d,%2d) : "
		   "mult=% 7.3f\n",
		   descriptors[j][0],descriptors[j][1],
		   descriptors[j][2],descriptors[j][3],
                   descriptors[j][4],descriptors[j][5],
		   multipliers[j]);
            if (n_out+1==MAX_out)
              printf("excessive output is stopped after %d terms\n",MAX_out);
          }
          n_out++;
          numoterms[descriptors[j][1]]++;
          hash[ hidx[j] ]=0; // reset hash entry
          hidx[j]=0; // not necessary
	  
	  // now overwrite the old stored information 
	  
	}
        for (i=1;i<=counter;i++) { hash[ hidx[i] ]=0; }
	counter=0;
      }
  }

  if (Control.output > 2 && myid == 0 && n_out < MAX_out)
    printf("IN:  Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)"
           " : mult=% 7.3f counter=%ld\n",
       indices[0],indices[1],
       indices[2],indices[3],indices[4],indices[5],dummy,counter);

  // the last config has lot of terms, try to reduce further by
  // reducing all possible permutations to one config by sorting (big first)

  for(;;) { // bubble sort indices
    j=1;
    for(i=2; i<5; i++)
      if (indices[i]<indices[i+1]) {  // exchange two indices
        j=0; k=indices[i+1]; indices[i+1]=indices[i]; indices[i]=k; 
      }
    if (j) break;
  }

  // Oct05: new hash table search (fast)
  j=0;
  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  {
    hval=(((((                          indices[5])
              *(NumberConfigurations+1)+indices[4])
              *(NumberConfigurations+1)+indices[3])
              *(NumberConfigurations+1)+indices[2])
              *(NumberConfigurations+1)+indices[0]);
    // try 5 following entries
    for(i=0;;i++) {
      hval = abs(hval + indices[2] + 1);  // avoid overflow to negativ values
      hval %= MAX_NUMBER_TERMS_IN_CCM_EQNS; // MAX should be prime
      j = hash[hval];
      if (j>counter || j<0)
         { printf("Bug86757 j=%ld hash=%d\n",j,hval); exit(1); }
      if (j==0) {      // empty entry
             hash[hval]=counter+1;  // set hash entry
             hidx[counter+1]=hval;  // set sequential index to entry
             j=counter+1;           // store to table
             hnew++; break;
      } else {
        if ( indices[0]==descriptors[j][0] &&
    term_13[indices[1]]==term_13[descriptors[j][1]] &&  // term
             indices[2]==descriptors[j][2] &&
             indices[3]==descriptors[j][3] &&
             indices[4]==descriptors[j][4] &&
             indices[5]==descriptors[j][5]
           ) {
            multipliers[j]+=dummy;  // hashed == searched
            hhit++; break;
         } else if (i>=1024) {
           hcol++; n_out++;    // collision, write through
           numoterms[indices[1]]++;
           term_store(id,indices,dummy);
           if (hcol<20) {
             printf("%2d: collision hash[%ld]=%ld lenhash=%d=%.3f%%\n",
                     myid, hval, hash[hval], counter,
                     counter*1.0/MAX_NUMBER_TERMS_IN_CCM_EQNS);
             printf("IN:   Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     indices[0],indices[1],
                     indices[2],indices[3],
                     indices[4],indices[5]);
             printf("HASH: Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     descriptors[j][0],descriptors[j][1],
                     descriptors[j][2],descriptors[j][3],
                     descriptors[j][4],descriptors[j][5]);
           }
           j=0; break;
         }
      }
    }
  }
  
  if (indices[0]>=0 && indices[0]<=NumberConfigurations) {
    if (indices[1]<0 || indices[1]>62) { 
      printf("Error collect L%d: term=%d\n",__LINE__,indices[1]); exit(1); }
    numiterms[indices[1]]++; // for statistics only
  }

  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  if ( j>counter /* && indices[1] != 0 */ ) {
    counter++;  /* potential overflow! */
    if (counter>max_counter) max_counter=counter; /* just as info */
    descriptors[counter][0]=indices[0];
    descriptors[counter][1]=indices[1];
    descriptors[counter][2]=indices[2];
    descriptors[counter][3]=indices[3];
    descriptors[counter][4]=indices[4];
    descriptors[counter][5]=indices[5];
    multipliers[counter]=dummy;
  }

  if (indices[0]>NumberConfigurations || indices[0]<0) {
    term_store(id,indices,dummy);

    if (Control.output > 1 && myid == 0 ) {
      printf("OUT: Config=%3d : Term = %2d\n", indices[0], indices[1]);
    }
    if (Control.output > 0 /* && myid == 0 */ ) {
      /* ToDo: clock output here? */
      int t2, tmin, tmax, tsum;
      t2=seconds(); // differ for each thread (collect min,max)
      tmin=tmax=tsum=t2;
#if (PARALLEL_FLAG> 0)
      int tmp[63], i;
      long tmp2[63];
      MPI_Allreduce(numiterms, tmp2, 63, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      for(i=0;i<63;i++) numiterms[i]=tmp2[i];   // input terms
      MPI_Allreduce(numoterms, tmp2, 63, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      for(i=0;i<63;i++) numoterms[i]=tmp2[i];   // output terms
      if (myid>0) for(i=0;i<63;i++) numoterms[i]=numiterms[i]=0;
      MPI_Allreduce(    &t2, &tmin,  1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tmax,  1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tsum,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      MPI_Allreduce(  &hcol,   tmp,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      hcol=tmp[0];
      MPI_Allreduce(  &n_in,  tmp2,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      n_in=tmp2[0];
      MPI_Allreduce( &n_out,  tmp2,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      n_out=tmp2[0];
      MPI_Allreduce(&max_counter,tmp2,  1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      max_counter=tmp2[0];
#endif
      /* output debugging infos */
      if (id_old != id) { id_old=id; na_out=na_in=0; } // new file
      na_out+=n_out; na_in+=n_in-1;
      // output this as a statistic information
      if (myid==0) {
        long sumi=0, sumo=0;
        for (i=0;i<63;i++){
          if(numiterms[i]) printf(" number of terms%02d : %8ld => %8ld\n",
                                  i, numiterms[i], numoterms[i]);
          sumi+=numiterms[i];
          sumo+=numoterms[i];
        }
        printf(" collect_term maxbuff= %8ld\n", max_counter);
        printf(" collect_term hash collisions= %6d\n", hcol);
        printf(" collect_term bond summed terms in= %8ld  out= %8ld"
               " t=%d..%d..%ds\n", sumi, sumo, tmin, tsum/npe, tmax);
        printf(" collect_term bond summed terms in= %8ld  out= %8ld\n",
              na_in, na_out);
      }
    }
    max_counter=0; n_out=n_in=0;
    hnew=hhit=hcol=0;   // reset number of collisions
  }
}
#else /* dont collect */
void collect_term(int indices[6],double dummy, int id) 
{
  static long n_out=0;
  int nio;

  term_store(id,indices,dummy);
	  
  if (Control.output > 1 && myid == 0 && n_out < MAX_out) {
     printf("OUT: Config=%d : Term = %d : CCM = (%d,%d,%d,%d) : "
     "mult=%10.7f\n",
     indices[0],indices[1],
     indices[2],indices[3],
     indices[4],indices[5],
     dummy);
     if (n_out+1==MAX_out)
        printf("excessive output is stopped after %d terms\n", MAX_out);
  }
  if (indices[0]>NumberConfigurations || indices[0]<0) {
    /* all terms collected, obtain finished for one bond */
#if (PARALLEL_FLAG> 0)
      long tmp;
      MPI_Allreduce(&na_out,  &tmp,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      na_out=tmp;
#endif
    if (Control.output > 0 && myid == 0 ) {
      /* output debugging infos */
      printf(" collect_term  terms out= %8ld\n", n_out);
      n_out=0;
    }
  }
  n_out++;
}
#endif
